/* $Id: misc.c,v 1.6 2002/10/17 19:47:46 haver Exp $ */

/*
 * Copyright (c) 2002 Srebrenko Sehic <haver@insecure.dk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */ 

#include <signal.h>
#include <stdio.h>
#include <termios.h>
#include <login_cap.h>
#include <bsd_auth.h>

#include "misc.h"

/*
 * Read passphrase if private key used for rpfcd is encrypted (it should be).
 */
char *
read_passphrase(const char *prompt)
{
	static char buf[MAX_PASSPHRASE_LEN + 1];
	char *ptr;
	int c;

	sigset_t sig, sigsave;
	struct termios term, termsave;
	FILE *fp;

	if ((fp = fopen(ctermid(NULL), "r+")) == NULL)
		return (NULL);

	setbuf(fp, NULL);

	sigemptyset(&sig);
	sigaddset(&sig, SIGINT);
	sigaddset(&sig, SIGTSTP);
	sigprocmask(SIG_BLOCK, &sig, &sigsave);

	tcgetattr(fileno(fp), &termsave);
	term = termsave;
	term.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
	tcsetattr(fileno(fp), TCSAFLUSH, &term);

	fputs(prompt, fp);

	ptr = buf;
	while ((c = getc(fp)) != EOF && c != '\n')
		if (ptr < &buf[MAX_PASSPHRASE_LEN])
			*ptr++ = c;

	*ptr = 0;
	putc('\n', fp);

	tcsetattr(fileno(fp), TCSAFLUSH, &termsave);
	sigprocmask(SIG_SETMASK, &sigsave, NULL);
	fclose(fp);

	return (buf);
}

/*
 * Return protocol name by number. We can't use getprotobynumber() since it
 * requires read access to /etc/protocols and we chroot to /var/empty before.
 */
char *
get_protocol_name(int number)
{
	switch (number) {
	case 1:
		return (PROTO_ICMP);
	case 6:
		return (PROTO_TCP);
	case 17:
		return (PROTO_UDP);
	default:
		return (NULL);
	}

	return (NULL);
}
