/* $Id: pf_ioctl.c,v 1.15 2002/10/17 19:47:46 haver Exp $ */

/*
 * Copyright (c) 2002 Srebrenko Sehic <haver@insecure.dk>
 * Copyright (c) 2001 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */ 
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/icmp6.h>
#include <arpa/inet.h>
#include <net/pfvar.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <netdb.h>
#include <stdarg.h>
#include <errno.h>
#include <err.h>

#include "misc.h"
#include "protocol.h"
#include "pf_ioctl.h"

char *tcpflags = "FSRPAUEW";

static const struct icmpcodeent icmp_code[] = {
	{ "net-unr",            ICMP_UNREACH,   ICMP_UNREACH_NET },
	{ "host-unr",           ICMP_UNREACH,   ICMP_UNREACH_HOST },
	{ "proto-unr",          ICMP_UNREACH,   ICMP_UNREACH_PROTOCOL },
	{ "port-unr",           ICMP_UNREACH,   ICMP_UNREACH_PORT },
	{ "needfrag",           ICMP_UNREACH,   ICMP_UNREACH_NEEDFRAG },
	{ "srcfail",            ICMP_UNREACH,   ICMP_UNREACH_SRCFAIL },
	{ "net-unk",            ICMP_UNREACH,   ICMP_UNREACH_NET_UNKNOWN },
	{ "host-unk",           ICMP_UNREACH,   ICMP_UNREACH_HOST_UNKNOWN },
	{ "isolate",            ICMP_UNREACH,   ICMP_UNREACH_ISOLATED },
	{ "net-prohib",         ICMP_UNREACH,   ICMP_UNREACH_NET_PROHIB },
	{ "host-prohib",        ICMP_UNREACH,   ICMP_UNREACH_HOST_PROHIB },
	{ "net-tos",            ICMP_UNREACH,   ICMP_UNREACH_TOSNET },
	{ "host-tos",           ICMP_UNREACH,   ICMP_UNREACH_TOSHOST },
	{ "filter-prohib",      ICMP_UNREACH,   ICMP_UNREACH_FILTER_PROHIB },
	{ "host-preced",        ICMP_UNREACH,   ICMP_UNREACH_HOST_PRECEDENCE },
	{ "cutoff-preced",      ICMP_UNREACH,   ICMP_UNREACH_PRECEDENCE_CUTOFF },
	{ "redir-net",          ICMP_REDIRECT,  ICMP_REDIRECT_NET },
	{ "redir-host",         ICMP_REDIRECT,  ICMP_REDIRECT_HOST },
	{ "redir-tos-net",      ICMP_REDIRECT,  ICMP_REDIRECT_TOSNET },
	{ "redir-tos-host",     ICMP_REDIRECT,  ICMP_REDIRECT_TOSHOST },
	{ "normal-adv",         ICMP_ROUTERADVERT, ICMP_ROUTERADVERT_NORMAL },
	{ "common-adv",         ICMP_ROUTERADVERT, ICMP_ROUTERADVERT_NOROUTE_COMMON },
	{ "transit",            ICMP_TIMXCEED,  ICMP_TIMXCEED_INTRANS },
	{ "reassemb",           ICMP_TIMXCEED,  ICMP_TIMXCEED_REASS },
	{ "badhead",            ICMP_PARAMPROB, ICMP_PARAMPROB_ERRATPTR },
	{ "optmiss",            ICMP_PARAMPROB, ICMP_PARAMPROB_OPTABSENT },
	{ "badlen",             ICMP_PARAMPROB, ICMP_PARAMPROB_LENGTH },
	{ "unknown-ind",        ICMP_PHOTURIS,  ICMP_PHOTURIS_UNKNOWN_INDEX },
	{ "auth-fail",          ICMP_PHOTURIS,  ICMP_PHOTURIS_AUTH_FAILED },
	{ "decrypt-fail",       ICMP_PHOTURIS,  ICMP_PHOTURIS_DECRYPT_FAILED }
};

static const struct icmpcodeent icmp6_code[] = {
	{ "admin-unr", ICMP6_DST_UNREACH, ICMP6_DST_UNREACH_ADMIN },
	{ "noroute-unr", ICMP6_DST_UNREACH, ICMP6_DST_UNREACH_NOROUTE },
	{ "notnbr-unr", ICMP6_DST_UNREACH, ICMP6_DST_UNREACH_NOTNEIGHBOR },
	{ "beyond-unr", ICMP6_DST_UNREACH, ICMP6_DST_UNREACH_BEYONDSCOPE },
	{ "addr-unr", ICMP6_DST_UNREACH, ICMP6_DST_UNREACH_ADDR },
	{ "port-unr", ICMP6_DST_UNREACH, ICMP6_DST_UNREACH_NOPORT },
	{ "transit", ICMP6_TIME_EXCEEDED, ICMP6_TIME_EXCEED_TRANSIT },
	{ "reassemb", ICMP6_TIME_EXCEEDED, ICMP6_TIME_EXCEED_REASSEMBLY },
	{ "badhead", ICMP6_PARAM_PROB, ICMP6_PARAMPROB_HEADER },
	{ "nxthdr", ICMP6_PARAM_PROB, ICMP6_PARAMPROB_NEXTHEADER },
	{ "redironlink", ND_REDIRECT, ND_REDIRECT_ONLINK },
	{ "redirrouter", ND_REDIRECT, ND_REDIRECT_ROUTER }
};

static const struct icmptypeent icmp_type[] = {
	{ "echoreq",    ICMP_ECHO },
	{ "echorep",    ICMP_ECHOREPLY },
	{ "unreach",    ICMP_UNREACH },
	{ "squench",    ICMP_SOURCEQUENCH },
	{ "redir",      ICMP_REDIRECT },
	{ "althost",    ICMP_ALTHOSTADDR },
	{ "routeradv",  ICMP_ROUTERADVERT },
	{ "routersol",  ICMP_ROUTERSOLICIT },
	{ "timex",      ICMP_TIMXCEED },
	{ "paramprob",  ICMP_PARAMPROB },
	{ "timereq",    ICMP_TSTAMP },
	{ "timerep",    ICMP_TSTAMPREPLY },
	{ "inforeq",    ICMP_IREQ },
	{ "inforep",    ICMP_IREQREPLY },
	{ "maskreq",    ICMP_MASKREQ },
	{ "maskrep",    ICMP_MASKREPLY },
	{ "trace",      ICMP_TRACEROUTE },
	{ "dataconv",   ICMP_DATACONVERR },
	{ "mobredir",   ICMP_MOBILE_REDIRECT },
	{ "ipv6-where", ICMP_IPV6_WHEREAREYOU },
	{ "ipv6-here",  ICMP_IPV6_IAMHERE },
	{ "mobregreq",  ICMP_MOBILE_REGREQUEST },
	{ "mobregrep",  ICMP_MOBILE_REGREPLY },
	{ "skip",       ICMP_SKIP },
	{ "photuris",   ICMP_PHOTURIS }
};

static const struct icmptypeent icmp6_type[] = {
	{ "unreach",    ICMP6_DST_UNREACH },
	{ "toobig",     ICMP6_PACKET_TOO_BIG },
	{ "timex",      ICMP6_TIME_EXCEEDED },
	{ "paramprob",  ICMP6_PARAM_PROB },
	{ "echoreq",    ICMP6_ECHO_REQUEST },
	{ "echorep",    ICMP6_ECHO_REPLY },
	{ "groupqry",   ICMP6_MEMBERSHIP_QUERY },
	{ "listqry",    MLD6_LISTENER_QUERY },
	{ "grouprep",   ICMP6_MEMBERSHIP_REPORT },
	{ "listenrep",  MLD6_LISTENER_REPORT },
	{ "groupterm",  ICMP6_MEMBERSHIP_REDUCTION },
	{ "listendone", MLD6_LISTENER_DONE },
	{ "routersol",  ND_ROUTER_SOLICIT },
	{ "routeradv",  ND_ROUTER_ADVERT },
	{ "neighbrsol", ND_NEIGHBOR_SOLICIT },
	{ "neighbradv", ND_NEIGHBOR_ADVERT },
	{ "redir",      ND_REDIRECT },
	{ "routrrenum", ICMP6_ROUTER_RENUMBERING },
	{ "wrureq",     ICMP6_WRUREQUEST },
	{ "wrurep",     ICMP6_WRUREPLY },
	{ "fqdnreq",    ICMP6_FQDN_QUERY },
	{ "fqdnrep",    ICMP6_FQDN_REPLY },
	{ "niqry",      ICMP6_NI_QUERY },
	{ "nirep",      ICMP6_NI_REPLY },
	{ "mtraceresp", MLD6_MTRACE_RESP },
	{ "mtrace",     MLD6_MTRACE }
};

const struct pf_timeout pf_timeouts[] = {
	{ "tcp.first",          PFTM_TCP_FIRST_PACKET },
	{ "tcp.opening",        PFTM_TCP_OPENING },
	{ "tcp.established",    PFTM_TCP_ESTABLISHED },
	{ "tcp.closing",        PFTM_TCP_CLOSING },
	{ "tcp.finwait",        PFTM_TCP_FIN_WAIT },
	{ "tcp.closed",         PFTM_TCP_CLOSED },
	{ "udp.first",          PFTM_UDP_FIRST_PACKET },
	{ "udp.single",         PFTM_UDP_SINGLE },
	{ "udp.multiple",       PFTM_UDP_MULTIPLE },
	{ "icmp.first",         PFTM_ICMP_FIRST_PACKET },
	{ "icmp.error",         PFTM_ICMP_ERROR_REPLY },
	{ "other.first",        PFTM_OTHER_FIRST_PACKET },
	{ "other.single",       PFTM_OTHER_SINGLE },
	{ "other.multiple",     PFTM_OTHER_MULTIPLE },
	{ "frag",               PFTM_FRAG },
	{ "interval",           PFTM_INTERVAL },
	{ NULL,                 0 }
};

/*
 * Query the kernel via ioctl(2) and get (DIOCGETSTATUS) pf statistics.
 * Values are put into a pf_status struct and returned to the caller.
 * See pf(4) for details.
 */
int
pf_get_stats(int dev, struct rpfcp_response *r_resp, int item)
{
	int i, len;
	time_t runtime;	
	struct pf_status pf_stats;
	char *ptr;
	char buf[RPFCP_MAX_MESSAGE_SIZE + 1];

	memset(buf, '\0', sizeof(buf));
	len = 0;
	ptr = buf;

	if (ioctl(dev, DIOCGETSTATUS, &pf_stats))
		return (PF_QUERY_FAILED);

	runtime = time(NULL) - pf_stats.since;

	if (pf_stats.running) {
		unsigned sec, min, hrs, day = runtime;
		
		switch (item) {
		case RPFCP_REQ_PF_UPTIME:
			sec = day % 60; day /= 60;
			min = day % 60; day /= 60;
			hrs = day % 24; day /= 24;

			snprintf(r_resp->msg, RPFCP_MESSAGE_LEN, "%u;%02u;%02u;%02u;", day, hrs, min, sec);
			return (PF_QUERY_OK);

		case RPFCP_REQ_PF_STATES:
			/* Loop thru states counter statistics and put it into a temp buffer. */
			for (i = 0; i < FCNT_MAX; i++)
				len = snprintf(ptr += len, RPFCP_MESSAGE_LEN, "%lld;", pf_stats.fcounters[i]);

			/* Now put the whole message together. */
			snprintf(r_resp->msg, RPFCP_MESSAGE_LEN, "%u;%s", pf_stats.states, buf);
			return (PF_QUERY_OK);
		case RPFCP_REQ_PF_COUNTERS:
			for (i = 0; i < PFRES_MAX; i++)
				len = snprintf(ptr += len, RPFCP_MESSAGE_LEN, "%lld;", pf_stats.counters[i]);

			snprintf(r_resp->msg, RPFCP_MESSAGE_LEN, "%s", buf);
			return (PF_QUERY_OK);
		default:
			return (-1);
		}
	}

	return (PF_DISABLED);
}

/*
 * Query internal pf(4) limits (states and frags).
 */
int
pf_get_limits(int dev, struct rpfcp_response *r_resp)
{
	int i, len;
	char *ptr;

	struct pfioc_limit pf_limit;

	ptr = r_resp->msg;

	/*
	 * We need to guery pf(4) twice since DIOCGETILMIT contains values for
	 * both number of 'states' and 'frags'.
	 */
	for (i = 0; i < 2; i++) {
		pf_limit.index = i;
		if (ioctl(dev, DIOCGETLIMIT, &pf_limit))
			return (PF_QUERY_FAILED);

		/* No limit is represented as 0. */
		if (pf_limit.limit == UINT_MAX)
			pf_limit.limit = 0;

		len = snprintf(ptr, RPFCP_MESSAGE_LEN, "%u;", pf_limit.limit);
		ptr += len;
	}

	return (PF_QUERY_OK);
}

/*
 * Query internal pf(4) timeout values.
 */
int
pf_get_timeouts(int dev, struct rpfcp_response *r_resp)
{
	int i, len;
	char *ptr;
	struct pfioc_tm pf_tm;

	ptr = r_resp->msg;

	/*
	 * pf keeps 16 different timeout values based on protocol and connection state.
	 * Hence, we itterate 16 times. There doesn't seem to be a constant defining this.
	 */
	for (i = 0; i < 16; i++) {
		pf_tm.timeout = i;
		if (ioctl(dev, DIOCGETTIMEOUT, &pf_tm))
			return (PF_QUERY_FAILED);

		len = snprintf(ptr, RPFCP_MESSAGE_LEN, "%d;", pf_tm.seconds);
		ptr += len;
	}

	return (PF_QUERY_OK);
}

/*
 * Return pf status (enabled or disabled).
 */
int
pf_get_status(int dev, struct rpfcp_response *r_resp)
{
	struct pf_status pf_stats;

	if (ioctl(dev, DIOCGETSTATUS, &pf_stats))
		return (PF_QUERY_FAILED);

	/* 1 = enabled; 0 = disabled */
	snprintf(r_resp->msg, RPFCP_MESSAGE_LEN, "%d;", pf_stats.running);

	return (PF_QUERY_OK);
}

/*
 * Open /dev/pf pseudo device and return the handle back to the non-privileged child.
 */
int
pf_open_device(void)
{
	int dev;

	dev = open(PF_DEVICE, O_RDONLY);
	return (dev == -1) ? 0 : dev;
}

/*
 * Close /dev/pf pseudo-device and return status of close(2) to the caller.
 */
int
pf_close_device(int dev)
{
	return (close(dev));
}

/*
 * Query the kernel for number of loaded rules.
 */
int
pf_get_number_of_filter_rules(int dev, int *rules_count)
{
	struct pfioc_rule pr;

	if (ioctl(dev, DIOCGETRULES, &pr))
		return (PF_QUERY_FAILED);

	*rules_count = pr.nr;	

	return (PF_QUERY_OK);
}

/*
 * Query all loaded rules in a syntax similar to output of pfctl -sr.
 */
int
pf_get_filter_rules(int dev, int rules, struct rpfcp_response *r_resp)
{
	struct pfioc_rule pr;
	char out[BUFSIZ];

	pr.nr = rules;
	if (ioctl(dev, DIOCGETRULE, &pr))
		return (PF_QUERY_FAILED);

	pf_print_rule(&pr.rule, out);
	out[strlen(out) - 1] = '\0';
	snprintf(r_resp->msg, RPFCP_MESSAGE_LEN, "%s;", out);

	return(PF_QUERY_OK);
}

/*
 * Format each rule and return the output buffer to pf_get_filter_rules().
 */
void
pf_print_rule(struct pf_rule *r, char *output)
{
	int i, opts, len;

	len = snprintf(output, BUFSIZ, "@%d ", r->nr); 
	if (r->action == PF_PASS)
		len = snprintf(output += len, BUFSIZ, "pass ");
	else if (r->action == PF_DROP) {
		len = snprintf(output += len, BUFSIZ, "block ");
		if (r->rule_flag & PFRULE_RETURNRST) {
			if (!r->return_ttl)
				len = snprintf(output += len, BUFSIZ, "return-rst ");
			else
				len = snprintf(output += len, BUFSIZ, "return-rst(ttl %d) ", r->return_ttl);
		} else if (r->return_icmp) {
			const struct icmpcodeent *ic;

			if (r->af != AF_INET6)
				len = snprintf(output += len, BUFSIZ, "return-icmp");
			else
				len = snprintf(output += len, BUFSIZ, "return-icmp6");
			ic = geticmpcodebynumber(r->return_icmp >> 8,
			    r->return_icmp & 255, r->af);

			if (ic == NULL)
				len = snprintf(output += len, BUFSIZ, "(%u) ", r->return_icmp & 255);
			else if ((r->af != AF_INET6 && ic->code !=
			    ICMP_UNREACH_PORT) ||
			    (r->af == AF_INET6 && ic->code !=
			    ICMP6_DST_UNREACH_NOPORT))
				len = snprintf(output += len, BUFSIZ, "(%s) ", ic->name);
			else
				len = snprintf(output += len, BUFSIZ, " ");
		}
	} else {
		len = snprintf(output += len, BUFSIZ, "scrub ");
	}
	if (r->direction == 0)
		len = snprintf(output += len, BUFSIZ, "in ");
	else
		len = snprintf(output += len, BUFSIZ, "out ");
	if (r->log == 1)
		len = snprintf(output += len, BUFSIZ, "log ");
	else if (r->log == 2)
		len = snprintf(output += len, BUFSIZ, "log-all ");
	if (r->quick)
		len = snprintf(output += len, BUFSIZ, "quick ");
	if (r->ifname[0]) {
		if (r->ifnot)
			len = snprintf(output += len, BUFSIZ, "on ! %s ", r->ifname);
		else
			len = snprintf(output += len, BUFSIZ, "on %s ", r->ifname);
	}
	if (r->rt) {
		if (r->rt == PF_ROUTETO)
			len = snprintf(output += len, BUFSIZ, "route-to ");
		else if (r->rt == PF_DUPTO)
			len = snprintf(output += len, BUFSIZ, "dup-to ");
		else if (r->rt == PF_FASTROUTE)
			len = snprintf(output += len, BUFSIZ, "fastroute");
		if (r->af && !PF_AZERO(&r->rt_addr, r->af)) {
			struct pf_addr_wrap aw;

			len = snprintf(output += len, BUFSIZ, "(%s ", r->rt_ifname);
			aw.addr = r->rt_addr;
			aw.addr_dyn = NULL;
			len = print_addr(&aw, NULL, r->af, output += len);
			len = snprintf(output += len, BUFSIZ, ")");
		} else if (r->rt_ifname[0])
			len = snprintf(output += len, BUFSIZ, "%s", r->rt_ifname);
		len = snprintf(output += len, BUFSIZ, " ");
	}
	if (r->af) {
		if (r->af == AF_INET)
			len = snprintf(output += len, BUFSIZ, "inet ");
		else
			len = snprintf(output += len, BUFSIZ, "inet6 ");
	}
	if (r->proto) {
		char *p_name = get_protocol_name(r->proto);

		if (p_name != NULL)
			len = snprintf(output += len, BUFSIZ, "proto %s ", p_name);
		else
			len = snprintf(output += len, BUFSIZ, "proto %u ", r->proto);
	}
	len = print_fromto(&r->src, &r->dst, r->af, r->proto, output += len);
	if (r->uid.op)
		len = print_uid(r->uid.op, r->uid.uid[0], r->uid.uid[1], "user", output += len);
	if (r->gid.op)
		len = print_gid(r->gid.op, r->gid.gid[0], r->gid.gid[1], "group", output += len);
	if (r->flags || r->flagset) {
		len = snprintf(output += len, BUFSIZ, "flags ");
		len = print_flags(r->flags, output += len);
		len = snprintf(output += len, BUFSIZ, "/");
		len = print_flags(r->flagset, output += len);
		len = snprintf(output += len, BUFSIZ, " ");
	}
	if (r->type) {
		const struct icmptypeent *p;

		p = geticmptypebynumber(r->type-1, r->af);
		if (r->af != AF_INET6)
			len = snprintf(output += len, BUFSIZ, "icmp-type");
		else
			len = snprintf(output += len, BUFSIZ, "ipv6-icmp-type");
		if (p != NULL)
			len = snprintf(output += len, BUFSIZ, " %s ", p->name);
		else
			len = snprintf(output += len, BUFSIZ, " %u ", r->type-1);
		if (r->code) {
			const struct icmpcodeent *p;

			p = geticmpcodebynumber(r->type-1, r->code-1, r->af);
			if (p != NULL)
				len = snprintf(output += len, BUFSIZ, "code %s ", p->name);
			else
				len = snprintf(output += len, BUFSIZ, "code %u ", r->code-1);
		}
	}
	if (r->keep_state == PF_STATE_NORMAL)
		len = snprintf(output += len, BUFSIZ, "keep state ");
	else if (r->keep_state == PF_STATE_MODULATE)
		len = snprintf(output += len, BUFSIZ, "modulate state ");
	opts = 0;
	if (r->max_states)
		opts = 1;
	for (i = 0; !opts && i < PFTM_MAX; ++i)
		if (r->timeout[i])
			opts = 1;
	if (opts) {
		len = snprintf(output += len, BUFSIZ, "(");
		if (r->max_states) {
			len = snprintf(output += len, BUFSIZ, "max %u", r->max_states);
			opts = 0;
		}
		for (i = 0; i < PFTM_MAX; ++i)
			if (r->timeout[i]) {
				if (!opts)
					len = snprintf(output += len, BUFSIZ, ", ");
				opts = 0;
				len = snprintf(output += len, BUFSIZ, "%s %u", pf_timeouts[i].name,
				    r->timeout[i]);
			}
		len = snprintf(output += len, BUFSIZ, ") ");
	}
	if (r->rule_flag & PFRULE_FRAGMENT)
		len = snprintf(output += len, BUFSIZ, "fragment ");
	if (r->rule_flag & PFRULE_NODF)
		len = snprintf(output += len, BUFSIZ, "no-df ");
	if (r->min_ttl)
		len = snprintf(output += len, BUFSIZ, "min-ttl %d ", r->min_ttl);
	if (r->max_mss)
		len = snprintf(output += len, BUFSIZ, "max-mss %d ", r->max_mss);
	if (r->allow_opts)
		len = snprintf(output += len, BUFSIZ, "allow-opts ");
	if (r->action == PF_SCRUB) {
		if (r->rule_flag & PFRULE_FRAGDROP)
			len = snprintf(output += len, BUFSIZ, "fragment drop-ovl ");
		else if (r->rule_flag & PFRULE_FRAGCROP)
			len = snprintf(output += len, BUFSIZ, "fragment crop ");
		else
			len = snprintf(output += len, BUFSIZ, "fragment reassemble ");
	}
	if (r->label[0])
		len = snprintf(output += len, BUFSIZ, "label %s", r->label);

	/*snprintf(output += len, BUFSIZ, "\n");*/
}

/*
 * Get UID and return to the caller. 
 */
int
print_uid(u_int8_t op, uid_t u1, uid_t u2, const char *t, char *output)
{
	int len;
	char a1[11], a2[11];

	snprintf(a1, sizeof(a1), "%u", u1);
	snprintf(a2, sizeof(a2), "%u", u2);
	len = snprintf(output, BUFSIZ, "%s ", t);
	if (u1 == UID_MAX && (op == PF_OP_EQ || op == PF_OP_NE))
		print_op(op, "unknown", a2, output += len);
	else
		print_op(op, a1, a2, output += len);

	return (len);
}

/*
 * Get GID and return to the caller.
 */
int
print_gid(u_int8_t op, gid_t g1, gid_t g2, const char *t, char *output)
{
	int len;
	char a1[11], a2[11];

	snprintf(a1, sizeof(a1), "%u", g1);
	snprintf(a2, sizeof(a2), "%u", g2);
	len = snprintf(output, BUFSIZ, "%s ", t);
	if (g1 == GID_MAX && (op == PF_OP_EQ || op == PF_OP_NE))
		print_op(op, "unknown", a2, output += len);
	else
		print_op(op, a1, a2, output += len);

	return (len);
}

/*
 * Get operator (used for port ranges) and return to the caller.
 */
int
print_op(u_int8_t op, const char *a1, const char *a2, char *output)
{
	int len;
	
	if (op == PF_OP_IRG)
		len = snprintf(output, BUFSIZ, "%s >< %s ", a1, a2);
	else if (op == PF_OP_XRG)
		len = snprintf(output, BUFSIZ, "%s <> %s ", a1, a2);
	else if (op == PF_OP_EQ)
		len = snprintf(output, BUFSIZ, "= %s ", a1);
	else if (op == PF_OP_NE)
		len = snprintf(output, BUFSIZ, "!= %s ", a1);
	else if (op == PF_OP_LT)
		len = snprintf(output, BUFSIZ, "< %s ", a1);
	else if (op == PF_OP_LE)
		len = snprintf(output, BUFSIZ, "<= %s ", a1);
	else if (op == PF_OP_GT)
		len = snprintf(output, BUFSIZ, "> %s ", a1);
	else if (op == PF_OP_GE)
		len = snprintf(output, BUFSIZ, ">= %s ", a1);

	return (len);
}

/*
 * Get port and return to the caller.
 */
int
print_port(u_int8_t op, u_int16_t p1, u_int16_t p2, char *proto, char *output)
{
	int len;
	char a1[6], a2[6];
	struct servent *s = getservbyport(p1, proto);

	p1 = ntohs(p1);
	p2 = ntohs(p2);
	snprintf(a1, sizeof(a1), "%u", p1);
	snprintf(a2, sizeof(a2), "%u", p2);
	len = snprintf(output, BUFSIZ, "port ");
	if (s != NULL && (op == PF_OP_EQ || op == PF_OP_NE))
		print_op(op, s->s_name, a2, output += len);
	else
		print_op(op, a1, a2, output += len);

	return (len);
}

/*
 * Get flags and return to the caller.
 */
int
print_flags(u_int8_t f, char *output)
{
	int i, len;

	for (i = 0; tcpflags[i]; ++i)
		if (f & (1 << i))
			len = snprintf(output, BUFSIZ, "%c", tcpflags[i]);

	return (len);
}

/*
 * Get IP address and return the caller.
 */
int
print_fromto(struct pf_rule_addr *src, struct pf_rule_addr *dst,
    u_int8_t af, u_int8_t proto, char *output)
{
	int len;
	
	if (PF_AZERO(&src->addr.addr, AF_INET6) &&
	    PF_AZERO(&src->mask, AF_INET6) &&
	    !src->noroute && !dst->noroute &&
	    !src->port_op && PF_AZERO(&dst->addr.addr, AF_INET6) &&
	    PF_AZERO(&dst->mask, AF_INET6) && !dst->port_op)
		len = snprintf(output, BUFSIZ, "all ");
	else {
		len = snprintf(output, BUFSIZ, "from ");
		if (src->noroute)
			len = snprintf(output += len, BUFSIZ, "no-route ");
		else if (PF_AZERO(&src->addr.addr, AF_INET6) &&
		    PF_AZERO(&src->mask, AF_INET6))
			len = snprintf(output += len, BUFSIZ, "any ");
		else {
			if (src->not)
				len = snprintf(output += len, BUFSIZ, "! ");
			len = print_addr(&src->addr, &src->mask, af, output += len);
			len = snprintf(output += len, BUFSIZ, " ");
		}
		if (src->port_op)
			print_port(src->port_op, src->port[0],
			    src->port[1],
			    proto == IPPROTO_TCP ? "tcp" : "udp", output += len);

		len = snprintf(output += len, BUFSIZ, "to ");
		if (dst->noroute)
			len = snprintf(output += len, BUFSIZ, "no-route ");
		else if (PF_AZERO(&dst->addr.addr, AF_INET6) &&
		    PF_AZERO(&dst->mask, AF_INET6))
			len = snprintf(output += len, BUFSIZ, "any ");
		else {
			if (dst->not)
				len = snprintf(output += len, BUFSIZ, "! ");
			len = print_addr(&dst->addr, &dst->mask, af, output += len);
			len = snprintf(output += len, BUFSIZ, " ");
		}
		if (dst->port_op)
			print_port(dst->port_op, dst->port[0],
			    dst->port[1],
			    proto == IPPROTO_TCP ? "tcp" : "udp", output += len);
	}

	return (len);
}

/*
 * Get IP address and return to the caller.
 */
int
print_addr(struct pf_addr_wrap *addr, struct pf_addr *mask, u_int8_t af, char *output)
{
	int len;
	char buf[48];
	len = 0;

	if (addr->addr_dyn != NULL)
		len = snprintf(output += len, BUFSIZ, "(%s)", addr->addr.pfa.ifname);
	else {
		if (inet_ntop(af, &addr->addr, buf, sizeof(buf)) == NULL)
			len += snprintf(output += len, BUFSIZ, "?");
		else
			len += snprintf(output += len, BUFSIZ, "%s", buf);
	}
	if (mask != NULL) {
		int bits = unmask(mask, af);

		if (bits != (af == AF_INET ? 32 : 128))
			len += snprintf(output += len, BUFSIZ, "/%u", bits);
	}

	return (len);
}

/*
 * Format netmask (/24, /16, /8) stuff.
 */
int
unmask(struct pf_addr *m, u_int8_t af)
{
    int i = 31, j = 0, b = 0, msize;
    u_int32_t tmp;

    if (af == AF_INET)
        msize = 1;
    else
        msize = 4;
    while (j < msize && m->addr32[j] == 0xffffffff) {
        b += 32;
        j++;
    }
    if (j < msize) {
        tmp = ntohl(m->addr32[j]);
        for (i = 31; tmp & (1 << i); --i)
            b++;
    }
    return (b);
}

const struct icmptypeent *
geticmptypebynumber(u_int8_t type, u_int8_t af)
{
    unsigned int i;

    if (af != AF_INET6) {
        for (i=0; i < (sizeof (icmp_type) / sizeof(icmp_type[0])); i++) {
            if (type == icmp_type[i].type)
                return (&icmp_type[i]);
        }
    } else {
        for (i=0; i < (sizeof (icmp6_type) /
            sizeof(icmp6_type[0])); i++) {
            if (type == icmp6_type[i].type)
                 return (&icmp6_type[i]);
        }
    }
    return (NULL);
}

const struct icmptypeent *
geticmptypebyname(char *w, u_int8_t af)
{
    unsigned int i;

    if (af != AF_INET6) {
        for (i=0; i < (sizeof (icmp_type) / sizeof(icmp_type[0])); i++) {
            if (!strcmp(w, icmp_type[i].name))
                return (&icmp_type[i]);
        }
    } else {
        for (i=0; i < (sizeof (icmp6_type) /
            sizeof(icmp6_type[0])); i++) {
            if (!strcmp(w, icmp6_type[i].name))
                return (&icmp6_type[i]);
        }
    }
    return (NULL);
}

const struct icmpcodeent *
geticmpcodebynumber(u_int8_t type, u_int8_t code, u_int8_t af)
{
    unsigned int i;

    if (af != AF_INET6) {
        for (i=0; i < (sizeof (icmp_code) / sizeof(icmp_code[0])); i++) {
            if (type == icmp_code[i].type &&
                code == icmp_code[i].code)
                return (&icmp_code[i]);
        }
    } else {
        for (i=0; i < (sizeof (icmp6_code) /
           sizeof(icmp6_code[0])); i++) {
            if (type == icmp6_code[i].type &&
                code == icmp6_code[i].code)
                return (&icmp6_code[i]);
        }
    }
    return (NULL);
}

const struct icmpcodeent *
geticmpcodebyname(u_long type, char *w, u_int8_t af)
{
    unsigned int i;

    if (af != AF_INET6) {
        for (i=0; i < (sizeof (icmp_code) / sizeof(icmp_code[0])); i++) {
            if (type == icmp_code[i].type &&
                !strcmp(w, icmp_code[i].name))
                return (&icmp_code[i]);
        }
    } else {
        for (i=0; i < (sizeof (icmp6_code) /
            sizeof(icmp6_code[0])); i++) {
            if (type == icmp6_code[i].type &&
                !strcmp(w, icmp6_code[i].name))
                return (&icmp6_code[i]);
        }
    }
    return (NULL);
}
