/* $Id: pf_ioctl.h,v 1.12 2002/10/17 13:16:40 haver Exp $ */

/*
 * Copyright (c) 2002 Srebrenko Sehic <haver@insecure.dk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */ 

#ifndef _PF_IOCTL_H_
#define _PF_IOCTL_H_

/*
 * Error values returned.
 */
#define PF_QUERY_OK				0
#define PF_QUERY_FAILED			1
#define PF_ACCESS_DENIED		2
#define PF_DISABLED				3

/*
 * PF pseudo device.
 */
#define PF_DEVICE	"/dev/pf"

/*
 * Function prototypes.
 */
int pf_get_stats(int, struct rpfcp_response *, int);
int pf_get_limits(int, struct rpfcp_response *);
int pf_get_timeouts(int, struct rpfcp_response *);
int pf_get_status(int, struct rpfcp_response *);
int pf_open_device(void);
int pf_close_device(int);
int pf_get_number_of_filter_rules(int, int *);
int pf_get_filter_rules(int, int, struct rpfcp_response *);
void pf_print_rule(struct pf_rule *, char *);
int print_op (u_int8_t, const char *, const char *, char *);
int print_port (u_int8_t, u_int16_t, u_int16_t, char *, char *);
int print_uid (u_int8_t, uid_t, uid_t, const char *, char *);
int print_gid (u_int8_t, gid_t, gid_t, const char *, char *);
int print_flags (u_int8_t, char *);
int print_fromto(struct pf_rule_addr *, struct pf_rule_addr *, u_int8_t, u_int8_t, char *);
int print_addr (struct pf_addr_wrap *, struct pf_addr *, u_int8_t, char *);
int unmask(struct pf_addr *m, u_int8_t af);

/*
 * PF struct definitions. 
 */
struct pf_timeout {
	const char  *name;
	int      timeout;
};

struct icmptypeent {
	char *name;
	u_int8_t type;
};

struct icmpcodeent {
	char *name;
	u_int8_t type;
	u_int8_t code;
};

const struct icmptypeent *geticmptypebynumber(u_int8_t, u_int8_t);
const struct icmptypeent *geticmptypebyname(char *, u_int8_t);
const struct icmpcodeent *geticmpcodebynumber(u_int8_t, u_int8_t, u_int8_t);
const struct icmpcodeent *geticmpcodebyname(u_long, char *, u_int8_t);
extern const struct pf_timeout pf_timeouts[];

#endif
