/* $Id: protocol.c,v 1.11 2002/10/17 10:03:58 haver Exp $ */

/*
 * Copyright (c) 2002 Srebrenko Sehic <haver@insecure.dk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */ 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include "protocol.h"

/*
 * Parse client request and check for RPFCP violations.
 */
int
parse_RPFCP_request(struct rpfcp_request *r_req, char *request)
{
	char *ptr;
	char rcode[4];
	int invalid = 0;
	int code = 0;
	size_t snaplen;

	ptr = request;

	/*
	 * Check for a valid header. Each message in RPFC protocol must
	 * begin with RPFCP_HEADER.
	 */
	if ((strncmp(ptr, RPFCP_HEADER, RPFCP_HEADER_LEN)) == 0) {
		strlcpy(r_req->header, ptr, RPFCP_HEADER_LEN);
		ptr += 5;
	}
	else
		invalid = 1;

	/* Check protcol major version. */
	if ((strncmp(ptr, "01", 2)) == 0) {
		r_req->major = 0;
		ptr += 2;
	}
	else
		invalid = 1;

	/* Check protocol minor version. */
	if ((strncmp(ptr, "00", 2)) == 0) {
		r_req->minor = 1;
		ptr += 2;
	}
	else
		invalid = 1;

	/* No protocol violations yet. Check request code. */
	strlcpy(rcode, ptr, sizeof(rcode));
	code = (int)strtol(rcode, (char **)NULL, 10);

	if ((valid_RPFCP_request_code(code)) < 1) {
		r_req->code = code;
		ptr += 3;
	}
	else
		invalid = 1;

	/*
	 * Copy message if any. Some request codes don't supply any arguments,
	 * while some might. E.g. username/password authentication.
	 */
	snaplen = strcspn(ptr, "\r\n");
	memcpy(r_req->msg, ptr, snaplen);
	ptr += snaplen;

	/* Check for terminating "\r\n". */
	if ((strncmp(ptr, "\r\n", 2)) != 0)
		invalid = 1;

	return (invalid) ? 1 : 0;
}

/*
 * Check for valid RPFCP request codes.
 */
int
valid_RPFCP_request_code(int code)
{
	return ((code >= RPFCP_REQUEST_LOW && code <= RPFCP_REQUEST_HIGH) ||
		(code == RPFCP_DISCONNECT || code == RPFCP_AUTH)) ? 0 : 1;
}

/*
 * Create a new RPFCP message based on the data received in the r_resp struct.
 */
void
make_RPFCP_response(struct rpfcp_response *r_resp, char *response)
{
	snprintf(response, RPFCP_MAX_MESSAGE_SIZE, "%s%02d%02d%03d%03d%s\r\n", 
		RPFCP_HEADER, RPFCP_MAJOR, RPFCP_MINOR, r_resp->code, r_resp->len, r_resp->msg);
}
