/* $Id: protocol.h,v 1.15 2002/10/17 13:16:40 haver Exp $ */

/*
 * Copyright (c) 2002 Srebrenko Sehic <haver@insecure.dk>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */ 

#ifndef _PROTOCOL_H_
#define _PROTOCOL_H_

/*
 * Maximum allowed size of messages in RPFC protocol.
 */
#define RPFCP_MAX_MESSAGE_SIZE		1024

/*
 * Protocol major and minor information.
 */
#define RPFCP_MAJOR    1
#define RPFCP_MINOR    0

/*
 * Header included in each request/response message.
 */
#define RPFCP_HEADER		"RPFCP"

/*
 * Length of various field(s) in the RPFC protocol.
 */
#define RPFCP_HEADER_LEN	5
#define RPFCP_MISC_LEN		10
#define RPFCP_META_LEN		RPFCP_HEADER_LEN + RPFCP_MISC_LEN
#define RPFCP_MESSAGE_LEN	RPFCP_MAX_MESSAGE_SIZE - RPFCP_META_LEN

/*
 * Request codes currently implemented in RPFC protocol.
 */
#define RPFCP_AUTH				101
#define RPFCP_DISCONNECT		102
#define RPFCP_REQ_PF_UPTIME		111
#define RPFCP_REQ_PF_STATES		112
#define RPFCP_REQ_PF_LIMITS		113
#define RPFCP_REQ_PF_COUNTERS	114
#define RPFCP_REQ_PF_TIMEOUTS	115
#define RPFCP_REQ_PF_STATUS		116
#define RPFCP_REQ_PF_FILTER		117

/*
 * Response codes currently implemented in RPFC protocol.
 */
#define RPFCP_NO_ERROR			300
#define RPFCP_MULTILINE_BEGIN	301
#define RPFCP_MULTILINE_END		302
#define RPFCP_VIOLATION			400
#define RPFCP_UNKNOWN_REQUEST	401
#define RPFCP_AUTH_FAILED		402
#define RPFCP_AUTH_REQUIRED		403
#define RPFCP_PF_QUERY_FAILED	500
#define RPFCP_PF_ACCESS_DENIED	501
#define RPFCP_PF_DISABLED		502

/*
 * Range for valid protocol request codes.
 */
#define RPFCP_REQUEST_LOW	111
#define RPFCP_REQUEST_HIGH	117

/*
 * Maximum length of username and password used in combination with
 * RPFCP_AUTH 101.
 */
#define MAX_USERNAME_LEN	16
#define MAX_PASSWORD_LEN	16

/*
 * RPFC protocol response and request structs.
 */
struct rpfcp_response {
    char header[RPFCP_HEADER_LEN + 1];
    int major;
    int minor;
    int code;
    int len;
    char msg[RPFCP_MESSAGE_LEN + 1];
};

struct rpfcp_request {
    char header[RPFCP_HEADER_LEN + 1];
    int major;
    int minor;
    int code;
    int len;
    char msg[RPFCP_MESSAGE_LEN + 1];
};

/*
 * Function prototypes.
 */
int parse_RPFCP_request(struct rpfcp_request *, char *);
int valid_RPFCP_request_code(int);
void make_RPFCP_response(struct rpfcp_response *, char *);

#endif
